"""Handle device info for legacy firmware so can offer upgrade."""

import re
from typing import Any


class Payload:
    """Handle device info for legacy firmware so can offer upgrade."""

    def __init__(self, data: dict[str, Any]) -> None:
        self.json: dict[str, Any] = data
        self.legacy_api: int = 1
        self.unpack()

    def unpack(self):
        """Unpack required info from the dashboard json."""
        if self.json:
            self.device_ip = self.json["ethIp"]
            self.MAC = self.json["wifiMac"]
            self.uptime = self.clean(self.json["uptime"])
            if "BUILD" in self.json:
                # legacy api = 1, old v2 builds < v2.3.6
                self.legacy_api = 1
                self.coord_mode = self.json["coordMode"]
                self.model = self.json["DEVICE"].replace("P", "p")
                self.sw_version = self.json["BUILD"]
                self.zb_hw = self.json["zbHw"]
                self.zb_version = self.clean(self.json["zbRev"])
            else:
                # legacy api = 2, 0.9.9 builds
                self.legacy_api = 2
                self.coord_mode = None
                self.model = self.json["hwRev"].replace("P", "p")
                self.sw_version = self.json["VERSION"].split(" ")[0]
                self.zb_hw = None
                self.zb_version = "-1"

    def clean(self, str: str) -> str:
        """
        Strip localisation formatting from string.

        Legacy firmware doesn't have required endpoints to get this directly.
        """
        text = None
        if "U_time" in str:
            text = str.replace("[U_time]^  ", "")
            text = text.replace("^", ":")
        elif "ZB_FW_unk" in str:
            text = str.replace("[ZB_FW_unk]", "-1")
        else:
            text = re.sub(r"\[|\]", "", str)
        return text



"""Exceptions for pysmlight"""


class SmlightError(Exception):
    """Generic SMLIGHT exception."""


class SmlightConnectionError(SmlightError):
    """SMLIGHT connection exception."""


class SmlightAuthError(SmlightError):
    """SMLIGHT authentication exception."""



from dataclasses import dataclass, field
import re

from mashumaro import DataClassDictMixin

from .payload import Payload


@dataclass
class Firmware(DataClassDictMixin):
    mode: str | None = None  # ESP|ZB
    type: int | None = None
    notes: str | None = None
    rev: str | None = None
    link: str | None = None
    ver: str | None = None
    dev: bool = False
    prod: bool = False
    baud: int | None = None

    def __post_init__(self):
        # zigbee firmware match key attributes
        if self.baud is not None:
            self.ver = self.rev
            self.dev = not self.prod

    def set_mode(self, mode):
        self.mode = mode


FirmwareList = list[Firmware] | None


@dataclass
class Radio(DataClassDictMixin):
    chip_index: int | None = None
    zb_channel: int | None = None
    zb_flash_size: int | None = None
    zb_hw: str | None = None
    zb_ram_size: int | None = None
    zb_version: str | None = None
    zb_type: int | None = None  # enum (coordinator, router, thread)
    radioModes: list[bool] | None = None


@dataclass
class Info(DataClassDictMixin):
    coord_mode: int | None = None  # Enum
    device_ip: str | None = None
    fs_total: int | None = None
    fw_channel: str | None = None  # dev, beta or release
    hostname: str | None = None
    legacy_api: int = 0
    MAC: str | None = None
    model: str | None = None
    ram_total: int | None = None
    sw_version: str | None = None
    wifi_mode: int | None = None  # enum (off, client, AP etc)
    radios: list[Radio] = field(default_factory=list)
    zb_channel: int | None = None
    zb_flash_size: int | None = None
    zb_hw: str | None = None
    zb_ram_size: int | None = None
    zb_version: str | None = None
    zb_type: int | None = None  # enum (coordinator, router, thread)

    @classmethod
    def load_payload(cls, payload: Payload) -> "Info":
        return cls(
            device_ip=payload.device_ip,
            legacy_api=payload.legacy_api,
            MAC=payload.MAC,
            model=payload.model,
            fw_channel="release",
            sw_version=payload.sw_version,
            zb_hw=payload.zb_hw,
            zb_version=int(payload.zb_version),
        )

    def check_zb_version(self, radio: Radio | None = None):
        if radio is None:
            radio = self

        if radio.zb_version is not None:
            if radio.zb_version == -1:
                radio.zb_channel = 2  # custom
                radio.zb_version = None
            else:
                radio.zb_version = str(radio.zb_version)
        return radio

    def __post_init__(self) -> None:
        if self.model is not None:
            self.model = self.model.replace("P", "p")
            self.model = self.model.replace("MG", "Mg")

        self.check_zb_version()

        # Factory firmware may have invalid .plus suffix, convert to valid version
        if self.sw_version:
            if "plus" in self.sw_version:
                self.sw_version = re.sub(
                    r"\.plus(\d*)$",
                    lambda m: f".{m.group(1) if m.group(1) else '1'}",
                    self.sw_version,
                )
            elif self.sw_version.endswith(".dev"):
                self.sw_version = f"{self.sw_version}0"

        if not self.radios:
            # construct radio object for backward compatibility (v2.5.0)
            self.radios = [
                Radio(
                    chip_index=0,
                    zb_channel=self.zb_channel,
                    zb_flash_size=self.zb_flash_size,
                    zb_hw=self.zb_hw,
                    zb_ram_size=self.zb_ram_size,
                    zb_version=self.zb_version,
                    zb_type=self.zb_type,
                )
            ]
        else:
            for r in self.radios:
                r = self.check_zb_version(r)


@dataclass
class Sensors(DataClassDictMixin):
    esp32_temp: float | None = None
    zb_temp: float | None = None
    zb_temp2: float | None = None
    uptime: int = 0  # Should be timestamp
    socket_uptime: int | None = None  # Should be timestamp
    ram_usage: int | None = None
    fs_used: int | None = None
    ethernet: bool = False
    wifi_connected: bool = False
    wifi_status: int | str | None = None  # enum (off, client, AP etc)
    vpn_status: bool = False

    # toggle states:
    disable_leds: bool | None = None
    night_mode: bool | None = None
    auto_zigbee: bool | None = None
    vpn_enabled: bool | None = None

    def __post_init__(self) -> None:
        if self.socket_uptime is not None and self.socket_uptime <= 0:
            self.socket_uptime = None


@dataclass
class SettingsEvent(DataClassDictMixin):
    page: int | None = None
    origin: str | None = None
    needReboot: bool = False
    setting: dict[str, bool | int] | None = None



from enum import Enum, IntEnum, auto, unique

FW_URL = "https://updates.smlight.tech/services/api/slzb-06x-ota.php"

MODE_LIST: dict[int, str] = {
    0: "eth",
    1: "wifi",
    2: "usb",
    3: "ppp",
}

PARAM_LIST: list[str] = [
    "coordMode",
    "zbRev",
    "espRev",
    "inetState",
    "locale",
    "newIpAvaiable",
]


@unique
class Actions(Enum):
    API_GET_PAGE = 0
    API_GET_PARAM = 1
    API_STARTWIFISCAN = 2
    API_SEND_HEX = 3
    API_CMD = 4
    API_GET_LOG = 5
    API_FLASH_ZB = 6
    API_WIFICONNECTSTAT = 7
    API_FLASH_ESP = 8
    API_ZHUB = 9


@unique
class Commands(Enum):
    CMD_ZB_ROUTER_RECON = 0
    CMD_ZB_RST = 1
    CMD_ZB_BSL = 2
    CMD_ESP_RES = 3
    CMD_CLEAR_LOG = 4
    CMD_ZB_ENERGY_SCAN = 5
    CMD_ZB_LED_NETWORK = 6
    CMD_ZB_LED_PERMIT = 7
    CMD_ZB_LED_DISABLED = 8
    CMD_HARD_RESET = 9
    CMD_TEMP_CALIB = 10
    CMD_ZB_IEEE_WRITE = 11
    CMD_ZB_IEEE_READ_FACTORY = 12
    CMD_WHTNW_MARK_READ = 13


@unique
class Pages(IntEnum):
    API2_PAGE_DASHBOARD = 0
    API2_PAGE_MODE = 1
    API2_PAGE_NETWORK = 2
    API2_PAGE_ZHA_Z2M = 3
    API2_PAGE_SECURITY = 4
    API2_PAGE_VPN = 5
    API2_PAGE_SETTINGS_GENERAL = 6
    API2_PAGE_SETTINGS_OTA = 7
    API2_PAGE_SETTINGS_LED = 8
    API2_PAGE_SETTINGS_TIME = 9
    API2_PAGE_SETTINGS_SYSTEM_LOG = 10
    API2_PAGE_ABOUT = 11
    API2_PAGE_FEEDBACK = 12
    API2_PAGE_ZHUB_DEVICES = 13
    API2_PAGE_ZHUB_DASHBOARD = 14


Devices: dict[str, int] = {
    "SLZB-06": 0,
    "SLZB-06M": 1,
    "SLZB-06Mg24": 2,
    "SLZB-06p7": 3,
    "SLZB-06p10": 4,
    "SLZB-06p7V2": 5,
    "SLZB-MR1": 6,
    "SLZB-MR2": 9,
    "SLZB-MR3": 10,
    "SLZB-MRW10": 11,
    "SLZB-06Mg26": 12,
    "SLZB-MR4": 13,
    "SLZB-MR1U": 14,
    "SLZB-06U": 15,
    "SMHUB-MG21": 65,
}


class U_Devices(Enum):
    """
    Represent devices based on ESP32-S3 core.

    Model id should match "Devices" dict.
    """

    SLZB_MR1U = 14
    SLZB_06U = 15


# Map radio's for MRx/U devices
MR_DEVICE_RADIO_MAP: dict[str, tuple[str, ...]] = {
    "SLZB-MR1": ("SLZB-06M", "SLZB-06p7V2"),
    "SLZB-MR2": ("SLZB-06M", "SLZB-06"),
    "SLZB-MR3": ("SLZB-06Mg24", "SLZB-06p10"),
    "SLZB-MR4": ("SLZB-06Mg26", "SLZB-06p10"),
    "SLZB-MR1U": ("SLZB-06M", "SLZB-06p7V2"),
    "SLZB-06U": ("SLZB-06",),
}


ZB_TYPES: dict[int, str] = {
    -1: "unknown",
    0: "coordinator",
    1: "router",
    2: "thread",
}

ZB_CHANNEL: dict[int, str] = {
    -1: "unknown",
    0: "stable",
    1: "dev",
    2: "custom",  # custom for 06M
}


class Events(Enum):
    API2_WIFISCANSTATUS = 0
    ESP_UPD_done = auto()
    EVENT_INET_STATE = auto()
    FW_UPD_done = auto()
    LOG_STR = auto()
    SAVE_PARAMS = auto()
    ZB_FW_err = auto()
    ZB_FW_info = auto()
    ZB_FW_prgs = auto()
    ZB_ENERGY_SCAN_DONE = auto()
    WHTNW = auto()
    REBOOT = auto()
    CATCH_ALL = 99


class Settings(Enum):
    DISABLE_LEDS = (Pages.API2_PAGE_SETTINGS_LED, "disableLeds")
    NIGHT_MODE = (Pages.API2_PAGE_SETTINGS_LED, "nightMode")
    ZB_AUTOUPDATE = (Pages.API2_PAGE_SETTINGS_OTA, "enabled")
    ENABLE_VPN = (Pages.API2_PAGE_VPN, "enabled")


class SettingsProp(Enum):
    DISABLE_LEDS = "disable_leds"
    NIGHT_MODE = "night_mode"
    ZB_AUTOUPDATE = "auto_zigbee"
    ENABLE_VPN = "vpn_enabled"


class WifiStatus(Enum):
    WL_NO_SHIELD = 255  # for compatibility with WiFi Shield library
    WL_IDLE_STATUS = 0
    WL_NO_SSID_AVAIL = 1
    WL_SCAN_COMPLETED = 2
    WL_CONNECTED = 3
    WL_CONNECT_FAILED = 4
    WL_CONNECTION_LOST = 5
    WL_DISCONNECTED = 6


class WifiMode(Enum):
    WIFI_MODE_NULL = 0  # null mode
    WIFI_MODE_STA = 1  # WiFi station mode
    WIFI_MODE_AP = 2  # WiFi soft-AP mode
    WIFI_MODE_APSTA = 3  # WiFi station + soft-AP mode
    WIFI_MODE_MAX = 4


class RebootReasons(IntEnum):
    ESP_RST_UNKNOWN = 0  # Reset reason can not be determined
    ESP_RST_POWERON = 1  # Reset due to power-on event
    ESP_RST_EXT = 2  # Reset by external pin (not applicable for ESP32)
    ESP_RST_SW = 3  # Software reset via esp_restart
    ESP_RST_PANIC = 4  # Software reset due to exception/panic
    ESP_RST_INT_WDT = 5  # Reset (software or hardware) due to interrupt watchdog
    ESP_RST_TASK_WDT = 6  # Reset due to task watchdog
    ESP_RST_WDT = 7  # Reset due to other watchdogs
    ESP_RST_DEEPSLEEP = 8  # Reset after exiting deep sleep mode
    ESP_RST_BROWNOUT = 9  # Brownout reset (software or hardware)
    ESP_RST_SDIO = 10  # Reset over SDIO



#!/usr/bin/env python3
from collections.abc import Callable
import json
import logging
import re
from typing import Any
import urllib.parse

from aiohttp import BasicAuth, ClientSession
from aiohttp.client_exceptions import ClientConnectionError
from awesomeversion import AwesomeVersion

from .const import (
    FW_URL,
    PARAM_LIST,
    Actions,
    Commands,
    Devices,
    Events,
    Pages,
    U_Devices,
)
from .exceptions import SmlightAuthError, SmlightConnectionError
from .models import Firmware, Info, Sensors
from .payload import Payload
from .sse import sseClient

_LOGGER = logging.getLogger(__name__)


class webClient:
    def __init__(self, host: str, session: ClientSession | None = None) -> None:
        self.auth: BasicAuth | None = None
        # we can't modify headers on the passed in session from HA,
        #  if needed can be overridden at request level
        self.headers = {"Content-Type": "application/json; charset=utf-8"}
        self.post_headers = {"Content-Type": "application/x-www-form-urlencoded"}
        self.host = host
        self.session = session
        self.close_session = False
        self.core_version: AwesomeVersion | None = None

        self.set_urls()

    async def authenticate(self, user: str, password: str) -> bool:
        """Pass in credentials and check auth is successful"""
        self.auth = BasicAuth(user, password)
        return not await self.check_auth_needed(True)

    async def check_auth_needed(self, authenticate: bool = False) -> bool:
        """
        Check if authentication is needed for the device
        Optionally validate authentication credentials
        Raises error on Connection or Auth failure
        """
        assert self.session is not None, "Session not created"

        auth: BasicAuth | None = None
        res = False
        if authenticate:
            auth = self.auth

        try:
            params = {"action": Actions.API_GET_PAGE.value, "page": 1}
            async with self.session.get(self.url, auth=auth, params=params) as response:
                if response.status == 401:
                    res = True
                    if authenticate:
                        raise SmlightAuthError("Authentication Error")
        except ClientConnectionError:
            _LOGGER.debug("Connection error")
            raise SmlightConnectionError("Connection failed")

        return res

    async def get(self, params: dict[str, Any], url: str | None = None) -> str | None:
        assert self.session is not None, "Session not created"

        if url is None:
            url = self.url

        try:
            async with self.session.get(
                url, headers=self.headers, params=params, auth=self.auth
            ) as response:
                if response.status == 404:
                    return None
                elif response.status == 401:
                    raise SmlightAuthError("Authentication Error")

                hdr = response.headers.get("respValuesArr")
                if hdr is not None and (
                    params and int(params["action"]) == Actions.API_GET_PAGE.value
                ):
                    return hdr
                else:
                    return await response.text(encoding="utf-8")
        except ClientConnectionError as err:
            raise SmlightConnectionError("Connection failed") from err

    async def post(self, params) -> bool:
        assert self.session is not None, "Session not created"

        data = urllib.parse.urlencode(params)

        try:
            async with self.session.post(
                self.setting_url,
                data=data,
                headers=self.post_headers,
                auth=self.auth,
            ) as response:
                if response.status == 404:
                    raise SmlightConnectionError("endpoint not found")
                elif response.status == 401:
                    raise SmlightAuthError("Authentication Error")
                await response.text(encoding="utf-8")
                return response.status == 200
        except ClientConnectionError as err:
            raise SmlightConnectionError("Connection failed") from err

    def set_host(self, host: str) -> None:
        self.host = host
        self.set_urls()

    def set_urls(self) -> None:
        self.url = f"http://{self.host}/api2"
        self.config_url = f"http://{self.host}/config"
        self.metrics_url = f"http://{self.host}/metrics"
        self.setting_url = f"http://{self.host}/settings/saveParams"
        self.info_url = f"http://{self.host}/ha_info"
        self.sensor_url = f"http://{self.host}/ha_sensors"

    async def close(self) -> None:
        """Close the session if it was created internally"""
        if self.session is not None and self.close_session:
            await self.session.close()
            self.session = None
            self.close_session = False

    async def __aenter__(self) -> "webClient":
        if self.session is None:
            self.close_session = True
            self.session = ClientSession(headers=self.headers)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: object | None,
    ) -> None:
        await self.close()


class Api2(webClient):
    def __init__(
        self,
        host: str,
        *,
        session: ClientSession | None = None,
        sse: sseClient | None = None,
    ) -> None:
        self.cmds = CmdWrapper(self.set_cmd)
        super().__init__(host, session=session)

        if session is None:
            self.session = ClientSession(headers=self.headers)
            self.close_session = True

        if sse:
            self.sse = sse
        else:
            self.sse = sseClient(host, session)

    async def get_device_payload(self) -> Payload:
        data = await self.get_page(Pages.API2_PAGE_DASHBOARD)
        res = Payload(data)
        return res

    async def get_firmware_version(
        self,
        channel: str | None,
        *,
        device: str | None = None,
        mode: str = "esp",
        zb_type: int | None = None,
        idx: int = 0,
    ) -> list[Firmware] | None:
        """Get firmware version for device and mode (esp | zigbee)"""
        params = {}
        fw_type = "ESP"
        if mode == "zigbee":
            fw_type = "ZB"
            params["format"] = "slzb"
            if device == "SLZB-MR1":
                device = "SLZB-06p7V2" if idx else "SLZB-06M"
        elif mode == "esp32":
            fw_type = "ESPs3" if self.device_is_u(device) else "ESP"
        params["type"] = fw_type
        response = await self.get(params=params, url=FW_URL)
        data = json.loads(response)

        if mode == "zigbee":
            assert device is not None
            data = data.get(str(Devices[device]), None)
        else:
            data = data["fw"]

        if data is None:
            return None

        fw = []
        for d in data:
            item = Firmware.from_dict(d)
            if not item.dev or channel == "dev":
                item.set_mode(fw_type)
                if item.notes:
                    item.notes = self.format_notes(item)
                if zb_type is not None and item.type != zb_type:
                    continue
                fw.append(item)
        return fw

    def format_notes(self, firmware: Firmware) -> str | None:
        """Format release notes for esp firmware"""
        if firmware and firmware.notes:
            items = (
                re.split("\r\n|(?<!\r)\n", firmware.notes)
                if firmware.mode == "ESP"
                else [firmware.notes]
            )
            notes = ""
            for i, v in enumerate(items):
                if i and v and not v.startswith("-"):
                    notes += f"* {v}\n"
                else:
                    notes += f"{v}\n\n"

            if firmware.dev and firmware.mode == "ZB":
                notes = "Dev firmware.\n\n" + notes
            return notes
        return None

    async def get_page(self, page: Pages) -> dict | None:
        """Extract Respvaluesarr json from page response header"""
        params = {"action": Actions.API_GET_PAGE.value, "page": page.value}
        res = await self.get(params)
        data = json.loads(res)
        return data if data else None

    async def get_param(self, param: str) -> str | None:
        if param in PARAM_LIST:
            params = {"action": Actions.API_GET_PARAM.value, "param": param}
            return await self.get(params)
        return None

    async def get_info_old(self) -> Info:
        self.sse.legacy_api = True
        payload = await self.get_device_payload()
        return Info.load_payload(payload)

    async def get_info(self) -> Info:
        res = await self.get(params=None, url=self.info_url)
        if res is None:
            return await self.get_info_old()
        elif res == "URL NOT FOUND":
            self.url = f"http://{self.host}/api"
            return await self.get_info_old()

        data = json.loads(res)

        info = Info.from_dict(data["Info"])
        core_version = AwesomeVersion(info.sw_version)

        if self.core_version is None:
            self.core_version = core_version
        if self.sse.sw_version is None:
            self.sse.sw_version = core_version

        return Info.from_dict(data["Info"])

    async def get_sensors(self) -> Sensors:
        res = await self.get(params=None, url=self.sensor_url)
        data = json.loads(res)
        return Sensors.from_dict(data["Sensors"])

    async def set_cmd(self, cmd: Commands, extra: str | None = None) -> bool:
        params = {"action": Actions.API_CMD.value, "cmd": cmd.value}
        if extra:
            k, v = extra.split(":")
            val = int(v)
            if val > 0:
                params[k] = val
        res = await self.get(params)
        return res == "ok"

    async def fw_update(
        self,
        firmware: Firmware,
        idx: int = 0,
    ) -> bool:
        """Send firmware update command to device"""
        if firmware.mode == "ZB":
            params = {
                "action": Actions.API_FLASH_ZB.value,
                "baud": firmware.baud,
                "fwUrl": firmware.link,
                "fwType": firmware.type,
                "fwVer": firmware.ver,
                "fwCh": int(not firmware.prod),
                "zbChipIdx": idx,
            }
            # backwards compatibility for SLZB-MR1
            if (
                idx == 1
                and self.core_version
                and self.core_version <= AwesomeVersion("v2.7.2")
            ):
                params["zbChipNum"] = 5
        else:
            params = {"action": Actions.API_FLASH_ESP.value, "fwUrl": firmware.link}
        res = await self.get(params)
        return res == "ok"

    async def set_toggle(self, page: Pages, toggle: str, value: bool) -> bool:
        state = "on" if value else "off"
        params = {"pageId": page.value, toggle: state, "ha": True}
        res = await self.post(params)
        return res

    async def scan_wifi(self, callback: Callable) -> Callable[[], None]:
        """Initiate scan of wifi networks.

        Args:
            callback (Callable): Callback function to process scan results

        Returns:
            Callable[[], None]: Function to clean up callback
        """

        remove_cb = self.sse.register_callback(Events.API2_WIFISCANSTATUS, callback)
        params = {"action": Actions.API_STARTWIFISCAN.value}
        await self.get(params)
        return remove_cb

    def device_is_u(self, model: str) -> bool:
        device_id = Devices.get(model, None)
        return (
            device_id in [udev.value for udev in U_Devices]
            if device_id is not None
            else False
        )


class CmdWrapper:
    """Convenience wrapper for HA when sending commands to the device."""

    def __init__(self, set_cmd: Callable) -> None:
        self.set_cmd = set_cmd

    async def reboot(self) -> None:
        await self.set_cmd(Commands.CMD_ESP_RES)

    async def zb_bootloader(self) -> None:
        await self.set_cmd(Commands.CMD_ZB_BSL)

    async def zb_restart(self) -> None:
        await self.set_cmd(Commands.CMD_ZB_RST)

    async def zb_router(self, idx: int = 0) -> None:
        await self.set_cmd(Commands.CMD_ZB_ROUTER_RECON, f"idx:{idx}")



__all__ = [
    "Api2",
    "CmdWrapper",
    "Firmware",
    "Info",
    "Radio",
    "Sensors",
    "SettingsEvent",
]

from pysmlight.models import Radio, SettingsEvent
from pysmlight.web import Api2, CmdWrapper, Firmware, Info, Sensors



"""Client to receive Server Sent Events (SSE) from the SMLIGHT API."""

import asyncio
from collections.abc import Callable
import json
import logging

import aiohttp
from aiohttp.client_exceptions import ClientConnectionError, SocketTimeoutError
from aiohttp_sse_client2.client import EventSource, MessageEvent
from awesomeversion import AwesomeVersion

from .const import Events, Pages, Settings
from .models import SettingsEvent

_LOGGER = logging.getLogger(__name__)

# override logging level aiohttp_sse_client2 library
aiologger = logging.getLogger("aiohttp_sse_client2")
aiologger.setLevel(logging.INFO)
LEGACY_SSE_VERSION = AwesomeVersion("v2.6.8.dev25")


class sseClient:
    """Initialise a client to receive Server Sent Events (SSE)"""

    def __init__(self, host: str, session: aiohttp.ClientSession):
        """Initialise the SSE client."""
        self.callbacks: dict[Events, Callable] = {}
        self.settings_cb: dict[Settings, Callable] = {}
        self.legacy_api = False
        self.sw_version: AwesomeVersion | None = None
        self.session = session
        self.url = f"http://{host}:81"  # Introduced in firmware v2.6.8.dev26
        self.legacy_url = f"http://{host}/events"
        self.setTimeout(30)

        self.register_callback(Events.SAVE_PARAMS, self._handle_settings)

    def setTimeout(self, timeout: int) -> None:
        self.timeout = aiohttp.ClientTimeout(
            total=None, connect=None, sock_connect=None, sock_read=timeout
        )

    async def client(self) -> None:
        """Run SSE Client

        SSE client will attempt reconnection on connection error with exponental backoff
        """

        # Increase timeout for legacy API which dont have PING events
        if self.legacy_api:
            self.setTimeout(600)

        while True:
            try:
                await self.sse_stream()
            except SocketTimeoutError:
                await asyncio.sleep(5)

    async def sse_stream(self) -> None:
        """Process incoming events on the message stream"""
        if self.sw_version is not None and self.sw_version <= LEGACY_SSE_VERSION:
            self.url = self.legacy_url
        async with EventSource(
            self.url, session=self.session, timeout=self.timeout, max_connect_retry=8
        ) as event_source:
            try:
                async for event in event_source:
                    _LOGGER.debug(event)
                    await self._message_handler(event)
            except (ClientConnectionError, ConnectionError) as err:
                _LOGGER.debug("Client Connection error: %s", err)
            else:
                _LOGGER.debug("Connection closed cleanly")

    async def _message_handler(self, event: MessageEvent) -> None:
        """Match event with callback for event type"""
        if event_type := getattr(Events, event.type, None):
            self.callbacks.get(event_type, lambda x: None)(event)
        self.callbacks.get(Events.CATCH_ALL, lambda x: None)(event)

    def register_callback(self, event: Events, cb: Callable) -> Callable[[], None]:
        """register a callback for a specific event type"""
        if event and event in self.callbacks:
            _LOGGER.warning("Callback for %s already exists, overwriting", event)

        self.callbacks[event] = cb

        def remove_callback():
            self.deregister_callback(event)

        return remove_callback

    def deregister_callback(self, event: Events) -> None:
        """Deregister callback for event type"""
        if event in self.callbacks:
            del self.callbacks[event]

    def _handle_settings(self, event: Events) -> None:
        """Process event and match callback for settings changes"""
        data = json.loads(event.data)
        try:
            page = Pages(data["page"])
        except ValueError:
            return

        changes = data.pop("changes", None)
        if changes is not None:
            for setting in changes:
                base = data.copy()
                match_cb = next(
                    (
                        cb
                        for k, cb in self.settings_cb.items()
                        if (page, setting) == k.value
                    ),
                    None,
                )

                if match_cb:
                    base["setting"] = {setting: changes[setting]}
                    result = SettingsEvent.from_dict(base)
                    match_cb(result)

    def register_settings_cb(
        self, setting: Settings, cb: Callable
    ) -> Callable[[], None]:
        """Register a callback for a specific setting"""
        self.settings_cb[setting] = cb

        def remove_callback():
            self.deregister_settings_cb(setting)

        return remove_callback

    def deregister_settings_cb(self, setting: Settings) -> None:
        """Deregister callback for a setting"""
        if setting in self.settings_cb:
            del self.settings_cb[setting]
